import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from functools import cached_property
from typing import ClassVar, Generic, Protocol, TypeVar, final

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.iam import User

logger = logging.getLogger(__name__)


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict]


Record = TypeVar("Record")


class AdministratorFinder(ABC):
    def __init__(self, ws: WorkspaceClient) -> None:
        self._ws = ws

    @abstractmethod
    def find_admin_users(self) -> Iterable[User]:
        """Locate active admin users."""
        raise NotImplementedError()


class WorkspaceAdministratorFinder(AdministratorFinder):
    """Locate the users that are in the 'admin' workspace group for a given workspace."""

    @staticmethod
    def _member_of_group_named(user: User, group_name: str) -> bool:
        """Determine whether a user belongs to a group with the given name or not."""
        return user.groups is not None and any(g.display == group_name for g in user.groups)

    @staticmethod
    def _member_of_group(user: User, group_id: str) -> bool:
        """Determine whether a user belongs to a group with the given identifier or not."""
        return user.groups is not None and any(g.value == group_id for g in user.groups)

    def _is_active_admin(self, user: User) -> bool:
        """Determine if a user is an active administrator."""
        return bool(user.active) and self._member_of_group_named(user, "admins")

    def _is_workspace_group(self, group_id: str) -> bool:
        """Determine whether a group_id corresponds to a workspace group or not."""
        try:
            group = self._ws.groups.get(group_id)
        except NotFound:
            return False
        return bool(group.meta and group.meta.resource_type == "WorkspaceGroup")

    def find_admin_users(self) -> Iterable[User]:
        """Enumerate the active workspace administrators in a given workspace.

        Returns:
            Iterable[User]: The active workspace administrators, if any.
        """
        logger.debug("Enumerating users to locate active workspace administrators...")
        all_users = self._ws.users.list(attributes="id,active,userName,groups")
        # The groups attribute is a flattened list of groups a user belongs to; hunt for the 'admins' workspace group.
        # Reference: https://learn.microsoft.com/en-us/azure/databricks/admin/users-groups/groups#account-vs-workspace-group
        admin_users = [user for user in all_users if user.user_name and self._is_active_admin(user)]
        logger.debug(f"Verifying membership of the 'admins' workspace group for users: {admin_users}")
        maybe_admins_id = set()
        for user in admin_users:
            if not user.groups:
                continue
            for group in user.groups:
                if group.display == "admins" and group.value:
                    maybe_admins_id.add(group.value)
        # There can only be a single 'admins' workspace group.
        for group_id in maybe_admins_id:
            if self._is_workspace_group(group_id):
                return (user for user in admin_users if self._member_of_group(user, group_id))
        return ()


class AccountAdministratorFinder(AdministratorFinder):
    """Locate the users that are account administrators for this workspace."""

    @staticmethod
    def _has_role(user: User, role: str) -> bool:
        """Determine whether a user has a given role or not."""
        return user.roles is not None and any(r.value == role for r in user.roles)

    def find_admin_users(self) -> Iterable[User]:
        """Enumerate the active account administrators associated with a given workspace.

        Returns:
            Iterable[User]: The active account administrators, if any.
        """
        logger.debug("Enumerating account users to locate active administrators...")
        response = self._ws.api_client.do(
            "GET", "/api/2.0/account/scim/v2/Users", query={"attributes": "id,active,userName,roles"}
        )
        assert isinstance(response, dict)
        all_users = (User.from_dict(resource) for resource in response.get("Resources", []))
        # Reference: https://learn.microsoft.com/en-us/azure/databricks/admin/users-groups/groups#account-admin
        return (user for user in all_users if user.active and user.user_name and self._has_role(user, "account_admin"))


class AdministratorLocator:
    """Locate a workspace administrator, if possible.

    This will first try to find an active workspace administrator. If there are multiple, the first (alphabetically
    sorted by user-name) will be used. If no active workspace administrators can be found then an account administrator
    is sought, again returning the first alphabetically by user-name if more than one is found.
    """

    def __init__(
        self,
        ws: WorkspaceClient,
        *,
        finders: Sequence[Callable[[WorkspaceClient], AdministratorFinder]] = (
            WorkspaceAdministratorFinder,
            AccountAdministratorFinder,
        ),
    ) -> None:
        """
        Initialize the instance, which will try to locate administrators using the workspace for the supplied client.

        Args:
            ws (WorkspaceClient): the client for workspace in which to locate admin users.
            finders: a sequence of factories that will be instantiated on demand to locate admin users.
        """
        self._ws = ws
        self._finders = finders

    @cached_property
    def _workspace_id(self) -> int:
        # Makes a REST call, so we cache it.
        return self._ws.get_workspace_id()

    @cached_property
    def _found_admin(self) -> str | None:

        # Ordering helper: User.user_name is typed as optional but we can't sort by None.
        # (The finders already filter out users without a user-name.)
        def _by_username(user: User) -> str:
            assert user.user_name
            return user.user_name

        # Lazily instantiate and query the finders in an attempt to locate an admin user.
        for factory in self._finders:
            finder = factory(self._ws)
            # First alphabetically by name.
            admin_user = min(finder.find_admin_users(), default=None, key=_by_username)
            if admin_user:
                return admin_user.user_name
        return None

    def get_workspace_administrator(self) -> str:
        """The user-name of an admin user for the workspace.

        Raises:
              RuntimeError if an admin user cannot be found in the current workspace.
        """
        found_admin = self._found_admin
        if found_admin is None:
            msg = f"No active workspace or account administrator can be found for workspace: {self._workspace_id}"
            raise RuntimeError(msg)
        return found_admin


class Ownership(ABC, Generic[Record]):
    """Determine an owner for a given type of object."""

    def __init__(self, administrator_locator: AdministratorLocator) -> None:
        self._administrator_locator = administrator_locator

    @final
    def owner_of(self, record: Record) -> str:
        """Obtain the user-name of a user that is responsible for the given record.

        This is intended to be a point of contact, and is either:

         - A user directly associated with the resource, such as the original creator; or
         - An active administrator for the current workspace.

        Args:
            record (Record): The record for which an associated user-name is sought.
        Returns:
            A string containing the user-name attribute of a user considered to be responsible for the resource.
        Raises:
            RuntimeError if there are no active administrators for the current workspace.
        """
        return self._maybe_direct_owner(record) or self._administrator_locator.get_workspace_administrator()

    @abstractmethod
    def _maybe_direct_owner(self, record: Record) -> str | None:
        """Obtain the record-specific user-name associated with the given record, if any."""
        return None



import dataclasses
import functools
import logging
import os
import re
import time
import webbrowser
from collections.abc import Callable, Iterable
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from typing import Any

from requests.exceptions import ConnectionError as RequestsConnectionError

import databricks.sdk.errors
from databricks.labs.blueprint.entrypoint import get_logger, is_in_debug
from databricks.labs.blueprint.installation import Installation, SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.parallel import ManyError, Threads
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.upgrades import Upgrades
from databricks.labs.blueprint.wheels import (
    ProductInfo,
    Version,
    find_project_root,
)
from databricks.labs.lsql.backends import SqlBackend, StatementExecutionBackend
from databricks.labs.lsql.dashboards import DashboardMetadata, Dashboards
from databricks.labs.lsql.deployment import SchemaDeployer
from databricks.sdk import WorkspaceClient, AccountClient
from databricks.sdk.useragent import with_extra
from databricks.sdk.errors import (
    AlreadyExists,
    BadRequest,
    DeadlineExceeded,
    InternalError,
    InvalidParameterValue,
    NotFound,
    PermissionDenied,
    ResourceAlreadyExists,
    ResourceDoesNotExist,
    Unauthenticated,
    OperationFailed,
)
from databricks.sdk.retries import retried
from databricks.sdk.service.dashboards import LifecycleState
from databricks.sdk.service.provisioning import Workspace
from databricks.sdk.service.sql import (
    CreateWarehouseRequestWarehouseType,
    EndpointInfoWarehouseType,
    SpotInstancePolicy,
)

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.assessment.azure import AzureServicePrincipalInfo
from databricks.labs.ucx.assessment.clusters import ClusterInfo, PolicyInfo
from databricks.labs.ucx.assessment.init_scripts import GlobalInitScriptInfo
from databricks.labs.ucx.assessment.jobs import JobInfo, SubmitRunInfo
from databricks.labs.ucx.assessment.pipelines import PipelineInfo
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.contexts.account_cli import AccountContext
from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext
from databricks.labs.ucx.framework.tasks import Task
from databricks.labs.ucx.hive_metastore.grants import Grant
from databricks.labs.ucx.hive_metastore.locations import ExternalLocation, Mount
from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatus
from databricks.labs.ucx.hive_metastore.table_size import TableSize
from databricks.labs.ucx.hive_metastore.tables import Table, TableError
from databricks.labs.ucx.hive_metastore.udfs import Udf
from databricks.labs.ucx.installer.hms_lineage import HiveMetastoreLineageEnabler
from databricks.labs.ucx.installer.logs import LogRecord
from databricks.labs.ucx.installer.mixins import InstallationMixin
from databricks.labs.ucx.installer.policy import ClusterPolicyInstaller
from databricks.labs.ucx.installer.workflows import WorkflowsDeployment
from databricks.labs.ucx.recon.migration_recon import ReconResult
from databricks.labs.ucx.runtime import Workflows
from databricks.labs.ucx.source_code.base import UsedTable
from databricks.labs.ucx.source_code.directfs_access import DirectFsAccess
from databricks.labs.ucx.source_code.jobs import JobProblem
from databricks.labs.ucx.source_code.queries import QueryProblem
from databricks.labs.ucx.workspace_access.base import Permissions
from databricks.labs.ucx.workspace_access.generic import WorkspaceObjectInfo
from databricks.labs.ucx.workspace_access.groups import ConfigureGroups, MigratedGroup

TAG_STEP = "step"
WAREHOUSE_PREFIX = "Unity Catalog Migration"
NUM_USER_ATTEMPTS = 10  # number of attempts user gets at answering a question

logger = logging.getLogger(__name__)


def deploy_schema(sql_backend: SqlBackend, inventory_schema: str):
    # we need to import it like this because we expect a module instance
    from databricks.labs import ucx  # pylint: disable=import-outside-toplevel

    logger.info("Creating ucx schemas...")
    deployer = SchemaDeployer(sql_backend, inventory_schema, ucx)
    deployer.deploy_schema()
    table = functools.partial(deployer.deploy_table)
    Threads.strict(
        "deploy tables",
        [
            functools.partial(table, "azure_service_principals", AzureServicePrincipalInfo),
            functools.partial(table, "clusters", ClusterInfo),
            functools.partial(table, "global_init_scripts", GlobalInitScriptInfo),
            functools.partial(table, "jobs", JobInfo),
            functools.partial(table, "pipelines", PipelineInfo),
            functools.partial(table, "external_locations", ExternalLocation),
            functools.partial(table, "mounts", Mount),
            functools.partial(table, "grants", Grant),
            functools.partial(table, "groups", MigratedGroup),
            functools.partial(table, "tables", Table),
            functools.partial(table, "table_size", TableSize),
            functools.partial(table, "table_failures", TableError),
            functools.partial(table, "workspace_objects", WorkspaceObjectInfo),
            functools.partial(table, "permissions", Permissions),
            functools.partial(table, "submit_runs", SubmitRunInfo),
            functools.partial(table, "policies", PolicyInfo),
            functools.partial(table, "migration_status", TableMigrationStatus),
            functools.partial(table, "workflow_problems", JobProblem),
            functools.partial(table, "query_problems", QueryProblem),
            functools.partial(table, "udfs", Udf),
            functools.partial(table, "logs", LogRecord),
            functools.partial(table, "recon_results", ReconResult),
            functools.partial(table, "directfs_in_paths", DirectFsAccess),
            functools.partial(table, "directfs_in_queries", DirectFsAccess),
            functools.partial(table, "used_tables_in_paths", UsedTable),
            functools.partial(table, "used_tables_in_queries", UsedTable),
        ],
    )
    deployer.deploy_view("grant_detail", "queries/views/grant_detail.sql")
    deployer.deploy_view("objects", "queries/views/objects.sql")
    deployer.deploy_view("table_estimates", "queries/views/table_estimates.sql")
    deployer.deploy_view("misc_patterns", "queries/views/misc_patterns.sql")
    deployer.deploy_view("code_patterns", "queries/views/code_patterns.sql")
    deployer.deploy_view("reconciliation_results", "queries/views/reconciliation_results.sql")
    deployer.deploy_view("directfs", "queries/views/directfs.sql")
    deployer.deploy_view("used_tables", "queries/views/used_tables.sql")


def extract_major_minor(version_string):
    match = re.search(r'(\d+\.\d+)', version_string)
    if match:
        return match.group(1)
    return None


class WorkspaceInstaller(WorkspaceContext):
    def __init__(
        self,
        ws: WorkspaceClient,
        environ: dict[str, str] | None = None,
        tasks: list[Task] | None = None,
    ):
        super().__init__(ws)
        if not environ:
            environ = dict(os.environ.items())
        self._force_install = environ.get("UCX_FORCE_INSTALL")
        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

        self._is_account_install = self._force_install == "account"
        self._tasks = tasks if tasks else Workflows.all().tasks()

    @cached_property
    def upgrades(self) -> Upgrades:
        return Upgrades(self.product_info, self.installation)

    @cached_property
    def policy_installer(self) -> ClusterPolicyInstaller:
        return ClusterPolicyInstaller(self.installation, self.workspace_client, self.prompts)

    @cached_property
    def installation(self) -> Installation:
        try:
            return self.product_info.current_installation(self.workspace_client)
        except NotFound:
            if self._force_install == "user":
                return Installation.assume_user_home(self.workspace_client, self.product_info.product_name())
            return Installation.assume_global(self.workspace_client, self.product_info.product_name())

    def run(
        self,
        default_config: WorkspaceConfig | None = None,
        config: WorkspaceConfig | None = None,
    ) -> WorkspaceConfig:
        logger.info(f"Installing UCX v{self.product_info.version()}")
        try:
            if config is None:
                config = self.configure(default_config)
            if self._is_testing():
                return config
            workflows_deployment = WorkflowsDeployment(
                config,
                self.installation,
                self.install_state,
                self.workspace_client,
                self.wheels,
                self.product_info,
                self._tasks,
            )
            workspace_installation = WorkspaceInstallation(
                config,
                self.installation,
                self.install_state,
                self.sql_backend,
                self.workspace_client,
                workflows_deployment,
                self.prompts,
                self.product_info,
            )
            workspace_installation.run()
        except ManyError as err:
            if len(err.errs) == 1:
                raise err.errs[0] from None
            raise err
        except TimeoutError as err:
            if isinstance(err.__cause__, RequestsConnectionError):
                logger.warning(
                    f"Cannot connect with {self.workspace_client.config.host} see "
                    f"https://github.com/databrickslabs/ucx#network-connectivity-issues for help: {err}"
                )
            raise err
        return config

    def _is_testing(self):
        return self.product_info.product_name() != "ucx"

    def _prompt_for_new_installation(self) -> WorkspaceConfig:
        logger.info("Please answer a couple of questions to configure Unity Catalog migration")
        default_database = "ucx"
        # if a workspace is configured to use external hive metastore, the majority of the time that metastore will be
        # shared with other workspaces. we need to add the suffix to ensure uniqueness of the inventory database
        if self.policy_installer.has_ext_hms():
            default_database = f"ucx_{self.workspace_client.get_workspace_id()}"
        inventory_database = self.prompts.question(
            "Inventory Database stored in hive_metastore", default=default_database, valid_regex=r"^\w+$"
        )
        ucx_catalog = self.prompts.question("Catalog to store UCX artifacts in", default="ucx", valid_regex=r"^\w+$")
        log_level = self.prompts.question("Log level", default="INFO").upper()
        num_threads = int(self.prompts.question("Number of threads", default="8", valid_number=True))
        configure_groups = ConfigureGroups(self.prompts)
        configure_groups.run()
        include_databases = self._select_databases()
        upload_dependencies = self.prompts.confirm(
            f"Does given workspace {self.workspace_client.get_workspace_id()} block Internet access?"
        )
        trigger_job = self.prompts.confirm("Do you want to trigger assessment job after installation?")
        recon_tolerance_percent = int(
            self.prompts.question("Reconciliation threshold, in percentage", default="5", valid_number=True)
        )
        return WorkspaceConfig(
            inventory_database=inventory_database,
            ucx_catalog=ucx_catalog,
            workspace_group_regex=configure_groups.workspace_group_regex,
            workspace_group_replace=configure_groups.workspace_group_replace,
            account_group_regex=configure_groups.account_group_regex,
            group_match_by_external_id=configure_groups.group_match_by_external_id,  # type: ignore[arg-type]
            include_group_names=configure_groups.include_group_names,
            renamed_group_prefix=configure_groups.renamed_group_prefix,
            log_level=log_level,
            num_threads=num_threads,
            include_databases=include_databases,
            trigger_job=trigger_job,
            recon_tolerance_percent=recon_tolerance_percent,
            upload_dependencies=upload_dependencies,
        )

    def _compare_remote_local_versions(self):
        try:
            local_version = self.product_info.released_version()
            remote_version = self.installation.load(Version).version
            if extract_major_minor(remote_version) == extract_major_minor(local_version):
                logger.info(f"UCX v{self.product_info.version()} is already installed on this workspace")
                msg = "Do you want to update the existing installation?"
                if not self.prompts.confirm(msg):
                    raise RuntimeWarning(
                        "UCX workspace remote and local install versions are same and no override is requested. Exiting..."
                    )
        except NotFound as err:
            logger.warning(f"UCX workspace remote version not found: {err}")

    def _confirm_force_install(self) -> bool:
        if not self._force_install:
            return False
        msg = "[ADVANCED] UCX is already installed on this workspace. Do you want to create a new installation?"
        if not self.prompts.confirm(msg):
            raise RuntimeWarning("UCX is already installed, but no confirmation")
        if not self.installation.is_global() and self._force_install == "global":
            # TODO:
            # Logic for forced global over user install
            # Migration logic will go here
            # verify complains without full path, asks to raise NotImplementedError builtin
            raise databricks.sdk.errors.NotImplemented("Migration needed. Not implemented yet.")
        if self.installation.is_global() and self._force_install == "user":
            # Logic for forced user install over global install
            self.replace(
                installation=Installation.assume_user_home(self.workspace_client, self.product_info.product_name())
            )
            return True
        return False

    def configure(self, default_config: WorkspaceConfig | None = None) -> WorkspaceConfig:
        """Configure the workspaces

        Notes:
        1. Connection errors are not handled within this configure method.
        """
        try:
            config = self.installation.load(WorkspaceConfig)
            self._compare_remote_local_versions()
            if self._confirm_force_install():
                return self._configure_new_installation(default_config)
            self._apply_upgrades()
            return config
        except NotFound as err:
            logger.debug(f"Cannot find previous installation: {err}")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            logger.warning(f"Existing installation at {self.installation.install_folder()} is corrupted. Skipping...")
        return self._configure_new_installation(default_config)

    def replace_config(self, **changes: Any) -> WorkspaceConfig | None:
        """
        Persist the list of workspaces where UCX is successfully installed in the config
        """
        try:
            config = self.installation.load(WorkspaceConfig)
            new_config = dataclasses.replace(config, **changes)
            self.installation.save(new_config)
        except (PermissionDenied, NotFound, ValueError):
            logger.warning(f"Failed to replace config for {self.workspace_client.config.host}")
            new_config = None
        return new_config

    def _apply_upgrades(self):
        try:
            self.upgrades.apply(self.workspace_client)
        except (InvalidParameterValue, NotFound) as err:
            logger.warning(f"Installed version is too old: {err}")

    def _configure_new_installation(self, default_config: WorkspaceConfig | None = None) -> WorkspaceConfig:
        if default_config is None:
            default_config = self._prompt_for_new_installation()
        HiveMetastoreLineageEnabler(self.workspace_client).apply(self.prompts, self._is_account_install)
        self._check_inventory_database_exists(default_config.inventory_database)
        warehouse_id = self.configure_warehouse()
        policy_id, instance_profile, spark_conf_dict, instance_pool_id = self.policy_installer.create(
            default_config.inventory_database
        )

        # Save configurable values for table migration cluster
        min_workers, max_workers, spark_conf_dict, managed_table_migration_choice = self._config_table_migration(
            spark_conf_dict
        )

        config = dataclasses.replace(
            default_config,
            warehouse_id=warehouse_id,
            instance_profile=instance_profile,
            spark_conf=spark_conf_dict,
            min_workers=min_workers,
            max_workers=max_workers,
            policy_id=policy_id,
            instance_pool_id=instance_pool_id,
            managed_table_external_storage=managed_table_migration_choice,
        )
        self.installation.save(config)
        if self._is_account_install:
            return config
        ws_file_url = self.installation.workspace_link(config.__file__)
        if self.prompts.confirm(f"Open config file in the browser and continue installing? {ws_file_url}"):
            webbrowser.open(ws_file_url)
        return config

    def _config_table_migration(self, spark_conf_dict) -> tuple[int, int, dict, str]:
        # parallelism will not be needed if backlog is fixed in https://databricks.atlassian.net/browse/ES-975874
        managed_table_migration_choices = {
            "Migrate MANAGED HMS table as EXTERNAL UC table. This option would require you to convert MANAGED HMS tables to EXTERNAL HMS tables once UC migration is complete, otherwise deleting HMS MANAGED table would delete the migrated UC table": 'SYNC_AS_EXTERNAL',
            "Copy data from MANAGED HMS to MANAGED UC table": 'CLONE',
        }
        managed_table_migration_choice = self.prompts.choice_from_dict(
            "If hive_metastore contains managed table with external"
            " location paths (not dbfs root), please select which migration "
            "option to choose from",
            managed_table_migration_choices,
        )
        if self._is_account_install:
            return 1, 10, spark_conf_dict, managed_table_migration_choice
        parallelism = self.prompts.question(
            "Parallelism for migrating dbfs root delta tables with deep clone", default="200", valid_number=True
        )
        if int(parallelism) > 200:
            spark_conf_dict.update({'spark.sql.sources.parallelPartitionDiscovery.parallelism': parallelism})
        # mix max workers for auto-scale migration job cluster
        min_workers = int(
            self.prompts.question(
                "Min workers for auto-scale job cluster for table migration", default="1", valid_number=True
            )
        )
        max_workers = int(
            self.prompts.question(
                "Max workers for auto-scale job cluster for table migration", default="10", valid_number=True
            )
        )

        return min_workers, max_workers, spark_conf_dict, managed_table_migration_choice

    def _select_databases(self):
        selected_databases = self.prompts.question(
            "Comma-separated list of databases to migrate. If not specified, we'll use all "
            "databases in hive_metastore",
            default="<ALL>",
        )
        if selected_databases != "<ALL>":
            return [x.strip() for x in selected_databases.split(",")]
        return None

    def configure_warehouse(self) -> str:
        def warehouse_type(_):
            return _.warehouse_type.value if not _.enable_serverless_compute else "SERVERLESS"

        pro_warehouses = {"[Create new PRO SQL warehouse]": "create_new"} | {
            f"{_.name} ({_.id}, {warehouse_type(_)}, {_.state.value})": _.id
            for _ in self.workspace_client.warehouses.list()
            if _.warehouse_type == EndpointInfoWarehouseType.PRO
        }
        if self._is_account_install:
            warehouse_id = "create_new"
        else:
            warehouse_id = self.prompts.choice_from_dict(
                "Select PRO or SERVERLESS SQL warehouse to run assessment dashboards on", pro_warehouses
            )
        if warehouse_id == "create_new":
            new_warehouse = self.workspace_client.warehouses.create(
                name=f"{WAREHOUSE_PREFIX} {time.time_ns()}",
                spot_instance_policy=SpotInstancePolicy.COST_OPTIMIZED,
                warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
                cluster_size="Small",
                max_num_clusters=1,
            )
            warehouse_id = new_warehouse.id
        return warehouse_id

    def _check_inventory_database_exists(self, inventory_database: str):
        logger.info("Fetching installations...")
        for installation in Installation.existing(self.workspace_client, self.product_info.product_name()):
            try:
                config = installation.load(WorkspaceConfig)
                if config.inventory_database == inventory_database:
                    raise AlreadyExists(
                        f"Inventory database '{inventory_database}' already exists in another installation"
                    )
            except (PermissionDenied, NotFound, SerdeError, ValueError, AttributeError):
                logger.warning(f"Existing installation at {installation.install_folder()} is corrupted. Skipping...")
                continue


class WorkspaceInstallation(InstallationMixin):
    def __init__(
        self,
        config: WorkspaceConfig,
        installation: Installation,
        install_state: InstallState,
        sql_backend: SqlBackend,
        ws: WorkspaceClient,
        workflows_installer: WorkflowsDeployment,
        prompts: Prompts,
        product_info: ProductInfo,
    ):
        self._config = config
        self._installation = installation
        self._install_state = install_state
        self._ws = ws
        self._sql_backend = sql_backend
        self._workflows_installer = workflows_installer
        self._prompts = prompts
        self._product_info = product_info
        environ = dict(os.environ.items())
        self._is_account_install = environ.get("UCX_FORCE_INSTALL") == "account"
        super().__init__(config, installation, ws)

    @classmethod
    def current(cls, ws: WorkspaceClient):
        # TODO: remove this method, it's no longer needed
        product_info = ProductInfo.from_class(WorkspaceConfig)
        installation = product_info.current_installation(ws)
        install_state = InstallState.from_installation(installation)
        config = installation.load(WorkspaceConfig)
        sql_backend = StatementExecutionBackend(ws, config.warehouse_id)
        wheels = product_info.wheels(ws)
        prompts = Prompts()
        tasks = Workflows.all().tasks()
        workflows_installer = WorkflowsDeployment(config, installation, install_state, ws, wheels, product_info, tasks)

        return cls(
            config,
            installation,
            install_state,
            sql_backend,
            ws,
            workflows_installer,
            prompts,
            product_info,
        )

    @property
    def config(self):
        return self._config

    @property
    def folder(self):
        return self._installation.install_folder()

    def run(self) -> bool:
        """Run workflow installation.

        Returns
            bool :
                True, installation finished. False installation did not finish.
        """
        logger.info(f"Installing UCX v{self._product_info.version()}")
        install_tasks = [self._create_database_and_dashboards, self._workflows_installer.create_jobs]
        Threads.strict("installing components", install_tasks)
        readme_url = self._installation.workspace_link("README")
        if not self._is_account_install and self._prompts.confirm(f"Open job overview in your browser? {readme_url}"):
            webbrowser.open(readme_url)
        logger.info(f"Installation completed successfully! Please refer to the {readme_url} for the next steps.")
        if self.config.trigger_job:
            logger.info("Triggering the assessment workflow")
            self._trigger_workflow("assessment")
        self._install_state.save()
        return True

    def _create_database_and_dashboards(self) -> None:
        self._create_database()  # Need the database before creating the dashboards
        Threads.strict("installing dashboards", list(self._get_create_dashboard_tasks()))

    # InternalError are retried for resilience on sporadic Databricks issues
    @retried(on=[InternalError], timeout=timedelta(minutes=2))
    def _create_database(self):
        try:
            deploy_schema(self._sql_backend, self._config.inventory_database)
        except Exception as err:
            if "UNRESOLVED_COLUMN.WITH_SUGGESTION" in str(err):
                msg = (
                    "The UCX version is not matching with the installed version."
                    "Kindly uninstall and reinstall UCX.\n"
                    "Please Follow the Below Command to uninstall and Install UCX\n"
                    "UCX Uninstall: databricks labs uninstall ucx.\n"
                    "UCX Install: databricks labs install ucx"
                )
                raise BadRequest(msg) from err
            if "Unable to load AWS credentials from any provider in the chain" in str(err):
                msg = (
                    "The UCX installation is configured to use external metastore. There is issue with the external metastore connectivity.\n"
                    "Please check the UCX installation instruction https://github.com/databrickslabs/ucx?tab=readme-ov-file#prerequisites"
                    "and re-run installation.\n"
                    "Please Follow the Below Command to uninstall and Install UCX\n"
                    "UCX Uninstall: databricks labs uninstall ucx.\n"
                    "UCX Install: databricks labs install ucx"
                )
                raise OperationFailed(msg) from err
            raise err

    def _get_create_dashboard_tasks(self) -> Iterable[Callable[[], None]]:
        """Get the tasks to create Lakeview dashboards from the SQL queries in the queries subfolders"""
        logger.info("Creating dashboards...")
        dashboard_folder_remote = f"{self._installation.install_folder()}/dashboards"
        try:
            self._ws.workspace.mkdirs(dashboard_folder_remote)
        except ResourceAlreadyExists:
            pass
        queries_folder = find_project_root(__file__) / "src/databricks/labs/ucx/queries"
        for step_folder in queries_folder.iterdir():
            if not step_folder.is_dir():
                continue
            logger.debug(f"Reading step folder {step_folder}...")
            for dashboard_folder in step_folder.iterdir():
                if not dashboard_folder.is_dir():
                    continue
                task = functools.partial(
                    self._create_dashboard,
                    dashboard_folder,
                    parent_path=dashboard_folder_remote,
                )
                yield task

    def _handle_existing_dashboard(self, dashboard_id: str, display_name: str, parent_path: str) -> str | None:
        """Handle an existing dashboard

        This method handles the following scenarios:
        - dashboard exists and needs to be updated
        - dashboard needs to be upgraded from Redash to Lakeview
        - dashboard is trashed and needs to be recreated
        - dashboard reference is invalid and the dashboard needs to be recreated

        Returns
            str | None :
                The dashboard id. If None, the dashboard will be recreated.
        """
        if "-" in dashboard_id:
            logger.info(f"Upgrading dashboard to Lakeview: {display_name} ({dashboard_id})")
            try:
                self._ws.dashboards.delete(dashboard_id=dashboard_id)
            except BadRequest:
                logger.warning(f"Cannot delete dashboard: {display_name} ({dashboard_id})")
            return None  # Recreate the dashboard if upgrading from Redash
        try:
            dashboard = self._ws.lakeview.get(dashboard_id)
            if dashboard.lifecycle_state is None:
                raise NotFound(f"Dashboard life cycle state: {display_name} ({dashboard_id})")
            if dashboard.lifecycle_state == LifecycleState.TRASHED:
                logger.info(f"Recreating trashed dashboard: {display_name} ({dashboard_id})")
                return None  # Recreate the dashboard if it is trashed (manually)
        except (NotFound, InvalidParameterValue):
            logger.info(f"Recovering invalid dashboard: {display_name} ({dashboard_id})")
            try:
                dashboard_path = f"{parent_path}/{display_name}.lvdash.json"
                self._ws.workspace.delete(dashboard_path)  # Cannot recreate dashboard if file still exists
                logger.debug(f"Deleted dangling dashboard {display_name} ({dashboard_id}): {dashboard_path}")
            except NotFound:
                pass
            return None  # Recreate the dashboard if it's reference is corrupted (manually)
        return dashboard_id  # Update the existing dashboard

    # InternalError and DeadlineExceeded are retried because of Lakeview internal issues
    # These issues have been reported to and are resolved by the Lakeview team
    # Keeping the retry for resilience
    @retried(on=[InternalError, DeadlineExceeded], timeout=timedelta(minutes=4))
    def _create_dashboard(self, folder: Path, *, parent_path: str) -> None:
        """Create a lakeview dashboard from the SQL queries in the folder"""
        logger.info(f"Creating dashboard in {folder}...")
        metadata = DashboardMetadata.from_path(folder).replace_database(
            database=f"hive_metastore.{self._config.inventory_database}",
            database_to_replace="inventory",
        )
        metadata.display_name = f"{self._name('UCX ')} {folder.parent.stem.title()} ({folder.stem.title()})"
        reference = f"{folder.parent.stem}_{folder.stem}".lower()
        dashboard_id = self._install_state.dashboards.get(reference)
        if dashboard_id is not None:
            dashboard_id = self._handle_existing_dashboard(dashboard_id, metadata.display_name, parent_path)
        dashboard = Dashboards(self._ws).create_dashboard(
            metadata,
            dashboard_id=dashboard_id,
            parent_path=parent_path,
            warehouse_id=self._warehouse_id,
            publish=True,
        )
        assert dashboard.dashboard_id is not None
        self._install_state.dashboards[reference] = dashboard.dashboard_id

    def uninstall(self):
        if self._prompts and not self._prompts.confirm(
            "Do you want to uninstall ucx from the workspace too, this would "
            "remove ucx project folder, dashboards, queries and jobs"
        ):
            return
        # TODO: this is incorrect, fetch the remote version (that appeared only in Feb 2024)
        logger.info(f"Deleting UCX v{self._product_info.version()} from {self._ws.config.host}")
        try:
            self._installation.files()
        except NotFound:
            logger.error(f"Check if {self._installation.install_folder()} is present")
            return
        self._check_and_fix_if_warehouse_does_not_exists()
        self._remove_database()
        self._remove_jobs()
        self._remove_warehouse()
        self._remove_policies()
        self._remove_secret_scope()
        self._installation.remove()
        logger.info("UnInstalling UCX complete")

    def _remove_database(self):
        if self._prompts and not self._prompts.confirm(
            f"Do you want to delete the inventory database {self._config.inventory_database} too?"
        ):
            return
        logger.info(f"Deleting inventory database {self._config.inventory_database}")
        deployer = SchemaDeployer(self._sql_backend, self._config.inventory_database, Any)
        deployer.delete_schema()

    def _remove_policies(self):
        logger.info("Deleting cluster policy")
        try:
            self._ws.cluster_policies.delete(policy_id=self.config.policy_id)
        except NotFound:
            logger.error("UCX Policy already deleted")

    def _remove_secret_scope(self):
        logger.info("Deleting secret scope")
        try:
            if self.config.uber_spn_id is not None:
                self._ws.secrets.delete_scope(self.config.inventory_database)
        except NotFound:
            logger.error("Secret scope already deleted")

    def _remove_jobs(self):
        self._workflows_installer.remove_jobs()

    def _remove_warehouse(self):
        try:
            warehouse_name = self._ws.warehouses.get(self._config.warehouse_id).name
            if warehouse_name.startswith(WAREHOUSE_PREFIX):
                logger.info(f"Deleting {warehouse_name}.")
                self._ws.warehouses.delete(id=self._config.warehouse_id)
        except InvalidParameterValue:
            logger.error("Error accessing warehouse details")

    def _trigger_workflow(self, step: str):
        job_id = int(self._install_state.jobs[step])
        job_url = f"{self._ws.config.host}#job/{job_id}"
        logger.debug(f"triggering {step} job: {self._ws.config.host}#job/{job_id}")
        self._ws.jobs.run_now(job_id)
        if self._prompts.confirm(f"Open {step} Job url that just triggered ? {job_url}"):
            webbrowser.open(job_url)

    def _check_and_fix_if_warehouse_does_not_exists(self):
        try:
            self._ws.warehouses.get(self._config.warehouse_id)
        except ResourceDoesNotExist:
            logger.critical(f"warehouse with id {self._config.warehouse_id} does not exists anymore")
            installer = WorkspaceInstaller(self._ws).replace(product_info=self._product_info, prompts=self._prompts)
            warehouse_id = installer.configure_warehouse()
            self._config = installer.replace_config(warehouse_id=warehouse_id)
            self._sql_backend = StatementExecutionBackend(self._ws, self._config.warehouse_id)


class AccountInstaller(AccountContext):
    def _get_safe_account_client(self) -> AccountClient:
        """
        Get account client with the correct host based on the cloud provider
        """
        if self.account_client.config.is_account_client:
            return self.account_client
        w = WorkspaceClient(product="ucx", product_version=__version__)
        host = w.config.environment.deployment_url("accounts")
        account_id = self.prompts.question("Please provide the Databricks account id")
        acct_client = AccountClient(host=host, account_id=account_id, product="ucx", product_version=__version__)
        self.replace(account_client=acct_client)
        return self.account_client

    def _get_installer(self, workspace: Workspace) -> WorkspaceInstaller:
        workspace_client = self.account_client.get_workspace_client(workspace)
        return WorkspaceInstaller(workspace_client).replace(product_info=self.product_info, prompts=self.prompts)

    def install_on_account(self):
        ctx = AccountContext(self._get_safe_account_client())
        default_config = None
        confirmed = False
        accessible_workspaces = self.account_workspaces.get_accessible_workspaces()
        msg = "\n".join([w.deployment_name for w in accessible_workspaces])
        installed_workspaces = []
        if not self.prompts.confirm(
            f"UCX has detected the following workspaces available to install. \n{msg}\nDo you want to continue?"
        ):
            return

        for workspace in accessible_workspaces:
            logger.info(f"Installing UCX on workspace {workspace.deployment_name}")
            installer = self._get_installer(workspace)
            if not confirmed:
                default_config = None
            try:
                default_config = installer.run(default_config)
            except RuntimeWarning:
                logger.info("Skipping workspace...")
            # if user confirms to install on remaining workspaces with the same config, don't prompt again
            installed_workspaces.append(workspace)
            if confirmed:
                continue
            confirmed = self.prompts.confirm(
                "Do you want to install UCX on the remaining workspaces with the same config?"
            )

        installed_workspace_ids = [w.workspace_id for w in installed_workspaces if w.workspace_id is not None]
        for workspace in installed_workspaces:
            installer = self._get_installer(workspace)
            installer.replace_config(installed_workspace_ids=installed_workspace_ids)

        # upload the json dump of workspace info in the .ucx folder
        ctx.account_workspaces.sync_workspace_info(installed_workspaces)

    def get_workspace_contexts(
        self, ws: WorkspaceClient, run_as_collection: bool, **named_parameters
    ) -> list[WorkspaceContext]:

        if not run_as_collection:
            return [WorkspaceContext(ws, named_parameters)]
        other_workspace_id = ws.get_workspace_id()
        workspace_contexts = []
        account_client = self._get_safe_account_client()
        acct_ctx = AccountContext(account_client)
        collection_workspace = account_client.workspaces.get(workspace_id=other_workspace_id)
        if not acct_ctx.account_workspaces.can_administer(collection_workspace):
            logger.error(f"User is not workspace admin of collection workspace {other_workspace_id}")
            return []
        installer = self._get_installer(collection_workspace)
        if installer.config.installed_workspace_ids is None:
            logger.error(f"No collection info found in the workspace {other_workspace_id}")
            return []
        for workspace_id in installer.config.installed_workspace_ids:
            workspace = account_client.workspaces.get(workspace_id)
            if not acct_ctx.account_workspaces.can_administer(workspace):
                logger.error(f"User is not workspace admin of workspace {workspace_id}")
                return []
            workspace_client = account_client.get_workspace_client(workspace)
            ctx = WorkspaceContext(workspace_client, named_parameters)
            workspace_contexts.append(ctx)
        return workspace_contexts

    @staticmethod
    def _validate_collection(workspace_ids: list[int], ids_to_workspace: dict[int, Workspace]) -> bool:
        # validate if the passed list of workspace ids, the user has workspace admin access on
        # if not, then cant join a collection
        for workspace_id in workspace_ids:
            if workspace_id not in ids_to_workspace:
                logger.error(
                    f"User doesnt have admin access on the workspace {workspace_id} in the collection, "
                    f"cant join collection."
                )
                return False
        return True

    def join_collection(self, workspace_ids: list[int], join_on_install: bool = False):
        ids_to_workspace: dict[int, Workspace] = {}
        if join_on_install:
            # if run as part of ucx installation prompt user if the installation needs to join a collection
            prompt_message = "Do you want to join the current installation to an existing collection?"
            if not self.prompts.confirm(prompt_message):
                return None

        account_client = self._get_safe_account_client()
        ctx = AccountContext(account_client)
        try:  # pylint: disable=too-many-try-statements
            # if user is account admin list all the available workspace the user has admin access on.
            # This code is run if joining collection after installation or through cli
            accessible_workspaces = ctx.account_workspaces.get_accessible_workspaces()
            for workspace in accessible_workspaces:
                if workspace.workspace_id is not None:
                    ids_to_workspace[workspace.workspace_id] = workspace
            if join_on_install:
                # if run as part of ucx installation allow user to select from the list to join
                target_workspace = self._get_collection_workspace(accessible_workspaces, account_client)
                assert target_workspace is not None and target_workspace.workspace_id is not None
                workspace_ids.append(target_workspace.workspace_id)
        except PermissionDenied:
            # if the user is not account admin, allow user to enter the workspace_id to join as collection.
            # if no workspace_id is entered, then exit
            logger.error("User doesnt have account admin permission, cant list workspaces")
            return None

        if not AccountInstaller._validate_collection(workspace_ids, ids_to_workspace):
            return None
        self._sync_collection(workspace_ids, ids_to_workspace)
        return None

    def _sync_collection(self, workspace_ids: list[int], ids_to_workspace: dict[int, Workspace]):
        # for all the workspace_ids syncs the config.installed_workspace_ids property in all the
        # collection workspaces
        installed_workspace_ids = []
        for workspace_id in workspace_ids:
            workspace = ids_to_workspace[workspace_id]
            installer = self._get_installer(workspace)
            workspace_collection_ids = installer.config.installed_workspace_ids
            if workspace_collection_ids is not None:
                installed_workspace_ids.extend(workspace_collection_ids)
            else:
                installed_workspace_ids.append(workspace_id)
        installed_workspace_ids = list(set(installed_workspace_ids))
        for workspace_id in installed_workspace_ids:
            installed_workspace = ids_to_workspace[workspace_id]
            installer = self._get_installer(installed_workspace)
            installer.replace_config(installed_workspace_ids=installed_workspace_ids)

    def _get_collection_workspace(
        self, accessible_workspaces: list[Workspace], account_client: AccountClient
    ) -> Workspace | None:
        # iterate through each workspace and select workspace which has existing ucx installation
        # allow user to select from the eligible workspace to join as collection
        installed_workspaces = []
        for workspace in accessible_workspaces:
            workspace_client = account_client.get_workspace_client(workspace)
            workspace_installation = Installation.existing(workspace_client, self.product_info.product_name())
            if len(workspace_installation) > 0:
                installed_workspaces.append(workspace)

        if len(installed_workspaces) == 0:
            logger.warning("No existing installation found , setting up new installation without")
            return None
        workspaces = {
            workspace.deployment_name: workspace
            for workspace in installed_workspaces
            if workspace.deployment_name is not None
        }
        target_workspace = self.prompts.choice_from_dict(
            "Please select a workspace, the current installation of ucx will be grouped as a "
            "collection with the selected workspace",
            workspaces,
        )
        return target_workspace


if __name__ == "__main__":
    with_extra("cmd", "install")
    logger = get_logger(__file__)
    if is_in_debug():
        logging.getLogger('databricks').setLevel(logging.DEBUG)
    env = dict(os.environ.items())
    force_install = env.get("UCX_FORCE_INSTALL")
    account_installer = AccountInstaller(AccountClient(product="ucx", product_version=__version__))
    if force_install == "account":
        account_installer.install_on_account()
    else:
        workspace_installer = WorkspaceInstaller(WorkspaceClient(product="ucx", product_version=__version__))
        workspace_installer.run()

        try:
            current_workspace_id = workspace_installer.workspace_client.get_workspace_id()
            account_installer.join_collection([current_workspace_id], True)
        except Unauthenticated:
            logger.warning(
                "User is not authenticated, "
                "you need account admin and workspace admin on "
                "respective workspaces to enable collection joining. Please run join-collection command "
                "with account admin credentials."
            )



from __future__ import annotations

import logging
from collections.abc import Sequence, Iterable

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import DatabricksError

from databricks.labs.ucx.framework.owners import Ownership
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.source_code.base import DirectFsAccess

logger = logging.getLogger(__name__)


class DirectFsAccessCrawler(CrawlerBase[DirectFsAccess]):

    @classmethod
    def for_paths(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_paths")

    @classmethod
    def for_queries(cls, backend: SqlBackend, schema) -> DirectFsAccessCrawler:
        return DirectFsAccessCrawler(backend, schema, "directfs_in_queries")

    def __init__(self, backend: SqlBackend, schema: str, table: str):
        """
        Initializes a DFSACrawler instance.

        Args:
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(backend=backend, catalog="hive_metastore", schema=schema, table=table, klass=DirectFsAccess)

    def dump_all(self, dfsas: Sequence[DirectFsAccess]):
        """This crawler doesn't follow the pull model because the fetcher fetches data for 2 crawlers, not just one
        It's not **bad** because all records are pushed at once.
        Providing a multi-entity crawler is out-of-scope of this PR
        """
        try:
            # TODO until we historize data, we append all DFSAs
            self._update_snapshot(dfsas, mode="append")
        except DatabricksError as e:
            logger.error("Failed to store DFSAs", exc_info=e)

    def _try_fetch(self) -> Iterable[DirectFsAccess]:
        sql = f"SELECT * FROM {escape_sql_identifier(self.full_name)}"
        for row in self._backend.fetch(sql):
            yield self._klass.from_dict(row.as_dict())

    def _crawl(self) -> Iterable[DirectFsAccess]:
        return []
        # TODO raise NotImplementedError() once CrawlerBase supports empty snapshots


class DirectFsAccessOwnership(Ownership[DirectFsAccess]):
    """Determine ownership of records reporting direct filesystem access.

    This is intended to be:

     - For queries, the creator of the query (if known).
     - For jobs, the owner of the path for the notebook or source (if known).

    At present this information is not gathered during the crawling process, so it can't be reported here.
    """

    def _maybe_direct_owner(self, record: DirectFsAccess) -> None:
        # TODO: Implement this once the creator/ownership information is exposed during crawling.
        return None



from typing import cast

from databricks.sdk.service.workspace import Language

from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
from databricks.labs.ucx.source_code.base import (
    Fixer,
    Linter,
    SqlSequentialLinter,
    CurrentSessionState,
    PythonSequentialLinter,
    PythonLinter,
    SqlLinter,
    TablePyCollector,
    TableSqlCollector,
    TableCollector,
    DfsaCollector,
    DfsaPyCollector,
    DfsaSqlCollector,
)
from databricks.labs.ucx.source_code.linters.directfs import DirectFsAccessPyLinter, DirectFsAccessSqlLinter
from databricks.labs.ucx.source_code.linters.imports import DbutilsPyLinter

from databricks.labs.ucx.source_code.linters.pyspark import (
    SparkSqlPyLinter,
    SparkTableNamePyLinter,
    SparkSqlTablePyCollector,
)
from databricks.labs.ucx.source_code.linters.spark_connect import SparkConnectPyLinter
from databricks.labs.ucx.source_code.linters.table_creation import DBRv8d0PyLinter
from databricks.labs.ucx.source_code.linters.from_table import FromTableSqlLinter


class LinterContext:
    def __init__(
        self,
        index: TableMigrationIndex | None = None,
        session_state: CurrentSessionState | None = None,
    ):
        self._index = index
        session_state = CurrentSessionState() if not session_state else session_state

        python_linters: list[PythonLinter] = []
        python_fixers: list[Fixer] = []
        python_dfsa_collectors: list[DfsaPyCollector] = []
        python_table_collectors: list[TablePyCollector] = []

        sql_linters: list[SqlLinter] = []
        sql_fixers: list[Fixer] = []
        sql_dfsa_collectors: list[DfsaSqlCollector] = []
        sql_table_collectors: list[TableSqlCollector] = []

        if index is not None:
            from_table = FromTableSqlLinter(index, session_state=session_state)
            sql_linters.append(from_table)
            sql_fixers.append(from_table)
            sql_table_collectors.append(from_table)
            spark_sql = SparkSqlPyLinter(from_table, from_table)
            python_linters.append(spark_sql)
            python_fixers.append(spark_sql)
            python_table_collectors.append(SparkSqlTablePyCollector(from_table))
            spark_table = SparkTableNamePyLinter(from_table, index, session_state)
            python_linters.append(spark_table)
            python_fixers.append(spark_table)
            python_table_collectors.append(spark_table)

        sql_direct_fs = DirectFsAccessSqlLinter()
        sql_linters.append(sql_direct_fs)
        sql_dfsa_collectors.append(sql_direct_fs)

        python_linters += [
            DirectFsAccessPyLinter(session_state),
            DBRv8d0PyLinter(dbr_version=session_state.dbr_version),
            SparkConnectPyLinter(session_state),
            DbutilsPyLinter(session_state),
            SparkSqlPyLinter(sql_direct_fs, None),
        ]

        python_dfsa_collectors += [DirectFsAccessPyLinter(session_state, prevent_spark_duplicates=False)]

        self._linters: dict[Language, list[SqlLinter] | list[PythonLinter]] = {
            Language.PYTHON: python_linters,
            Language.SQL: sql_linters,
        }
        self._fixers: dict[Language, list[Fixer]] = {
            Language.PYTHON: python_fixers,
            Language.SQL: sql_fixers,
        }

        self._dfsa_collectors: dict[Language, list[DfsaSqlCollector] | list[DfsaPyCollector]] = {
            Language.PYTHON: python_dfsa_collectors,
            Language.SQL: sql_dfsa_collectors,
        }

        self._table_collectors: dict[Language, list[TableSqlCollector] | list[TablePyCollector]] = {
            Language.PYTHON: python_table_collectors,
            Language.SQL: sql_table_collectors,
        }

    def is_supported(self, language: Language) -> bool:
        return language in self._linters and language in self._fixers

    def linter(self, language: Language) -> Linter:
        if language not in self._linters:
            raise ValueError(f"Unsupported language: {language}")
        if language is Language.PYTHON:
            return PythonSequentialLinter(cast(list[PythonLinter], self._linters[language]), [], [])
        if language is Language.SQL:
            return SqlSequentialLinter(cast(list[SqlLinter], self._linters[language]), [], [])
        raise ValueError(f"Unsupported language: {language}")

    def fixer(self, language: Language, diagnostic_code: str) -> Fixer | None:
        if language not in self._fixers:
            return None
        for fixer in self._fixers[language]:
            if fixer.name == diagnostic_code:
                return fixer
        return None

    def dfsa_collector(self, language: Language) -> DfsaCollector:
        if language not in self._dfsa_collectors:
            raise ValueError(f"Unsupported language: {language}")
        if language is Language.PYTHON:
            return PythonSequentialLinter([], cast(list[DfsaPyCollector], self._dfsa_collectors[language]), [])
        if language is Language.SQL:
            return SqlSequentialLinter([], cast(list[DfsaSqlCollector], self._dfsa_collectors[language]), [])
        raise ValueError(f"Unsupported language: {language}")

    def tables_collector(self, language: Language) -> TableCollector:
        if language not in self._table_collectors:
            raise ValueError(f"Unsupported language: {language}")
        if language is Language.PYTHON:
            return PythonSequentialLinter([], [], cast(list[TablePyCollector], self._table_collectors[language]))
        if language is Language.SQL:
            return SqlSequentialLinter([], [], cast(list[TableSqlCollector], self._table_collectors[language]))
        raise ValueError(f"Unsupported language: {language}")

    def apply_fixes(self, language: Language, code: str) -> str:
        linter = self.linter(language)
        for advice in linter.lint(code):
            fixer = self.fixer(language, advice.code)
            if fixer:
                code = fixer.apply(code)
        return code



import datetime as dt
import logging
from dataclasses import dataclass

from databricks.labs.lsql.backends import SqlBackend
from databricks.labs.lsql.deployment import SchemaDeployer

from databricks.labs.ucx.__about__ import __version__
from databricks.labs.ucx.hive_metastore.verification import MetastoreNotFoundError, VerifyHasCatalog, VerifyHasMetastore
from databricks.labs.ucx.installer.workflows import DeployedWorkflows
from databricks.labs.ucx.progress.workflow_runs import WorkflowRun


logger = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class Historical:
    workspace_id: int
    """The identifier of the workspace where this historical record was generated."""

    job_run_id: int
    """The identifier of the job run that generated this historical record."""

    object_type: str
    """The inventory table for which this historical record was generated."""

    object_id: list[str]
    """The type-specific identifier for the corresponding inventory record."""

    data: dict[str, str]
    """Type-specific JSON-encoded data of the inventory record."""

    failures: list[str]
    """The list of problems associated with the object that this inventory record covers."""

    owner: str
    """The identity that has ownership of the object."""

    ucx_version: str = __version__
    """The UCX semantic version."""


class ProgressTrackingInstallation:
    """Install resources for UCX's progress tracking."""

    _SCHEMA = "multiworkspace"

    def __init__(self, sql_backend: SqlBackend, ucx_catalog: str) -> None:
        # `mod` is a required parameter, though, it's not used in this context without views.
        self._schema_deployer = SchemaDeployer(sql_backend, self._SCHEMA, mod=None, catalog=ucx_catalog)

    def run(self) -> None:
        self._schema_deployer.deploy_schema()
        self._schema_deployer.deploy_table("workflow_runs", WorkflowRun)
        self._schema_deployer.deploy_table("historical", Historical)
        logger.info("Installation completed successfully!")


class VerifyProgressTracking:
    """Verify the progress tracking is ready to be used."""

    def __init__(
        self,
        verify_has_metastore: VerifyHasMetastore,
        verify_has_ucx_catalog: VerifyHasCatalog,
        deployed_workflows: DeployedWorkflows,
    ) -> None:
        self._verify_has_metastore = verify_has_metastore
        self._verify_has_ucx_catalog = verify_has_ucx_catalog
        self._deployed_workflows = deployed_workflows

    def verify(self, timeout=dt.timedelta(seconds=0)) -> None:
        """Verify the progress tracking installation is ready to be used.

        Prerequisites:
        - UC metastore exists.
        - UCX catalog exists.
        - A job run corresponding to the "assessment" job:
            - Finished successfully.
            - OR if pending or running, we will wait up to the timeout for the assessment run to finish. If it did still
              not finish successfully, we raise an error.

        Otherwise, we consider the prerequisites to be NOT matched.

        Args :
            timeout (datetime.timedelta) : Timeout to wait for pending or running assessment run.

        Raises :
            RuntimeWarning : Signalling the prerequisites are not met.
        """
        metastore_not_attached_message = (
            "Metastore not attached to workspace. Run `databricks labs ucx assign-metastore`"
        )
        try:
            has_metastore = self._verify_has_metastore.verify_metastore()
        except MetastoreNotFoundError as e:
            raise RuntimeWarning(metastore_not_attached_message) from e
        if not has_metastore:
            raise RuntimeWarning(metastore_not_attached_message)
        if not self._verify_has_ucx_catalog.verify():
            raise RuntimeWarning("UCX catalog not configured. Run `databricks labs ucx create-ucx-catalog`")
        if not self._deployed_workflows.validate_step("assessment", timeout=timeout):
            raise RuntimeWarning(
                "Assessment workflow did not complete successfully yet. "
                "Run `databricks labs ucx ensure-assessment-run` command"
            )



from __future__ import annotations

import builtins
import sys
from abc import ABC
import logging
import re
from collections.abc import Iterable
from typing import TypeVar, cast

from astroid import (  # type: ignore
    Assign,
    AssignName,
    Attribute,
    Call,
    ClassDef,
    Const,
    Expr,
    Import,
    ImportFrom,
    Instance,
    Module,
    Name,
    NodeNG,
    parse,
    Uninferable,
)

logger = logging.getLogger(__name__)
missing_handlers: set[str] = set()


T = TypeVar("T", bound=NodeNG)


class Tree:

    @staticmethod
    def parse(code: str):
        root = parse(code)
        return Tree(root)

    @classmethod
    def normalize_and_parse(cls, code: str):
        code = cls.normalize(code)
        root = parse(code)
        return Tree(root)

    @classmethod
    def normalize(cls, code: str):
        code = cls._normalize_indents(code)
        code = cls._convert_magic_lines_to_magic_commands(code)
        return code

    @classmethod
    def _normalize_indents(cls, python_code: str):
        lines = python_code.split("\n")
        for line in lines:
            # skip leading ws and comments
            if len(line.strip()) == 0 or line.startswith('#'):
                continue
            if not line.startswith(' '):
                # first line of code is correctly indented
                return python_code
            # first line of code is indented when it shouldn't
            prefix_count = len(line) - len(line.lstrip(' '))
            prefix_str = ' ' * prefix_count
            for i, line_to_fix in enumerate(lines):
                if line_to_fix.startswith(prefix_str):
                    lines[i] = line_to_fix[prefix_count:]
            return "\n".join(lines)
        return python_code

    @classmethod
    def _convert_magic_lines_to_magic_commands(cls, python_code: str):
        lines = python_code.split("\n")
        magic_markers = {"%", "!"}
        in_multi_line_comment = False
        pattern = re.compile('"""')
        for i, line in enumerate(lines):
            if len(line) == 0:
                continue
            if not in_multi_line_comment and line[0] in magic_markers:
                lines[i] = f"magic_command({line.encode()!r})"
                continue
            matches = re.findall(pattern, line)
            if len(matches) & 1:
                in_multi_line_comment = not in_multi_line_comment
        return "\n".join(lines)

    @classmethod
    def new_module(cls):
        node = Module("root")
        return Tree(node)

    def __init__(self, node: NodeNG):
        self._node: NodeNG = node

    @property
    def node(self):
        return self._node

    @property
    def root(self):
        node = self._node
        while node.parent:
            node = node.parent
        return node

    def walk(self) -> Iterable[NodeNG]:
        yield from self._walk(self._node)

    @classmethod
    def _walk(cls, node: NodeNG) -> Iterable[NodeNG]:
        yield node
        for child in node.get_children():
            yield from cls._walk(child)

    def locate(self, node_type: type[T], match_nodes: list[tuple[str, type]]) -> list[T]:
        visitor = MatchingVisitor(node_type, match_nodes)
        visitor.visit(self._node)
        return visitor.matched_nodes

    def first_statement(self):
        if isinstance(self._node, Module):
            if len(self._node.body) > 0:
                return self._node.body[0]
        return None

    def __repr__(self):
        truncate_after = 32
        code = repr(self._node)
        if len(code) > truncate_after:
            code = code[0:truncate_after] + "..."
        return f"<Tree: {code}>"

    def append_tree(self, tree: Tree) -> Tree:
        """returns the appended tree, not the consolidated one!"""
        if not isinstance(tree.node, Module):
            raise NotImplementedError(f"Can't append tree from {type(tree.node).__name__}")
        tree_module: Module = cast(Module, tree.node)
        self.append_nodes(tree_module.body)
        self.append_globals(tree_module.globals)
        # the following may seem strange but it's actually ok to use the original module as tree root
        # because each node points to the correct parent (practically, the tree is now only a list of statements)
        return tree

    def append_globals(self, globs: dict[str, list[Expr]]) -> None:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't append globals to {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        for name, values in globs.items():
            statements: list[Expr] = self_module.globals.get(name, None)
            if statements is None:
                self_module.globals[name] = list(values)  # clone the source list to avoid side-effects
                continue
            statements.extend(values)

    def append_nodes(self, nodes: list[NodeNG]) -> None:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't append statements to {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        for node in nodes:
            node.parent = self_module
            self_module.body.append(node)

    def is_instance_of(self, class_name: str) -> bool:
        for inferred in self.node.inferred():
            if inferred is Uninferable:
                continue
            if not isinstance(inferred, (Const, Instance)):
                return False
            proxied = getattr(inferred, "_proxied", None)
            return isinstance(proxied, ClassDef) and proxied.name == class_name
        return False

    def is_from_module(self, module_name: str) -> bool:
        return self._is_from_module(module_name, set())

    def _is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        if self._node in visited:
            logger.debug(f"Recursion encountered while traversing node {self._node.as_string()}")
            return False
        visited.add(self._node)
        return self._node_is_from_module(module_name, visited)

    def _node_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        if isinstance(self._node, Name):
            return self._name_is_from_module(module_name, visited)
        if isinstance(self._node, Call):
            return self._call_is_from_module(module_name, visited)
        if isinstance(self._node, Attribute):
            return self._attribute_is_from_module(module_name, visited)
        if isinstance(self._node, Const):
            return self._const_is_from_module(module_name, visited)
        return False

    def _name_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Name)
        # if this is the call's root node, check it against the required module
        if self._node.name == module_name:
            return True
        root = self.root
        if not isinstance(root, Module):
            return False
        for value in root.globals.get(self._node.name, []):
            if not isinstance(value, AssignName) or not isinstance(value.parent, Assign):
                continue
            if _LocalTree(value.parent.value).is_from_module_visited(module_name, visited):
                return True
        return False

    def _call_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Call)
        # walk up intermediate calls such as spark.range(...)
        return isinstance(self._node.func, Attribute) and _LocalTree(self._node.func.expr).is_from_module_visited(
            module_name, visited
        )

    def _attribute_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Attribute)
        return _LocalTree(self._node.expr).is_from_module_visited(module_name, visited)

    def _const_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Const)
        return _LocalTree(self._node.parent).is_from_module_visited(module_name, visited)

    def has_global(self, name: str) -> bool:
        if not isinstance(self.node, Module):
            return False
        self_module: Module = cast(Module, self.node)
        return self_module.globals.get(name, None) is not None

    def nodes_between(self, first_line: int, last_line: int) -> list[NodeNG]:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't extract nodes from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        nodes: list[NodeNG] = []
        for node in self_module.body:
            if node.lineno < first_line:
                continue
            if node.lineno > last_line:
                break
            nodes.append(node)
        return nodes

    def globals_between(self, first_line: int, last_line: int) -> dict[str, list[NodeNG]]:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't extract globals from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        globs: dict[str, list[NodeNG]] = {}
        for key, nodes in self_module.globals.items():
            nodes_in_scope: list[NodeNG] = []
            for node in nodes:
                if node.lineno < first_line or node.lineno > last_line:
                    continue
                nodes_in_scope.append(node)
            if len(nodes_in_scope) > 0:
                globs[key] = nodes_in_scope
        return globs

    def line_count(self) -> int:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't count lines from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        nodes_count = len(self_module.body)
        if nodes_count == 0:
            return 0
        return 1 + self_module.body[nodes_count - 1].lineno - self_module.body[0].lineno

    def renumber(self, start: int) -> Tree:
        assert start != 0
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't renumber {type(self.node).__name__}")
        root: Module = cast(Module, self.node)
        # for now renumber in place to avoid the complexity of rebuilding the tree with clones

        def renumber_node(node: NodeNG, offset: int) -> None:
            for child in node.get_children():
                renumber_node(child, offset + (child.lineno or 0) - (node.lineno or 0))
            if node.end_lineno:
                node.end_lineno = node.end_lineno + offset
                node.lineno = node.lineno + offset

        nodes = root.body if start > 0 else reversed(root.body)
        for node in nodes:
            offset = start - node.lineno
            renumber_node(node, offset)
            num_lines = 1 + (node.end_lineno - node.lineno if node.end_lineno else 0)
            start = start + num_lines if start > 0 else start - num_lines
        return self

    def is_builtin(self) -> bool:
        if isinstance(self._node, Name):
            name = self._node.name
            return name in dir(builtins) or name in sys.stdlib_module_names or name in sys.builtin_module_names
        if isinstance(self._node, Call):
            return Tree(self._node.func).is_builtin()
        if isinstance(self._node, Attribute):
            return Tree(self._node.expr).is_builtin()
        return False  # not supported yet


class _LocalTree(Tree):

    def is_from_module_visited(self, name: str, visited_nodes: set[NodeNG]) -> bool:
        return self._is_from_module(name, visited_nodes)


class TreeHelper(ABC):

    @classmethod
    def get_call_name(cls, call: Call) -> str:
        if not isinstance(call, Call):
            return ""
        func = call.func
        if isinstance(func, Name):
            return func.name
        if isinstance(func, Attribute):
            return func.attrname
        return ""  # not supported yet

    @classmethod
    def extract_call_by_name(cls, call: Call, name: str) -> Call | None:
        """Given a call-chain, extract its sub-call by method name (if it has one)"""
        assert isinstance(call, Call)
        node = call
        while True:
            func = node.func
            if not isinstance(func, Attribute):
                return None
            if func.attrname == name:
                return node
            if not isinstance(func.expr, Call):
                return None
            node = func.expr

    @classmethod
    def args_count(cls, node: Call) -> int:
        """Count the number of arguments (positionals + keywords)"""
        assert isinstance(node, Call)
        return len(node.args) + len(node.keywords)

    @classmethod
    def get_arg(
        cls,
        node: Call,
        arg_index: int | None,
        arg_name: str | None,
    ) -> NodeNG | None:
        """Extract the call argument identified by an optional position or name (if it has one)"""
        assert isinstance(node, Call)
        if arg_index is not None and len(node.args) > arg_index:
            return node.args[arg_index]
        if arg_name is not None:
            arg = [kw.value for kw in node.keywords if kw.arg == arg_name]
            if len(arg) == 1:
                return arg[0]
        return None

    @classmethod
    def is_none(cls, node: NodeNG) -> bool:
        """Check if the given AST expression is the None constant"""
        if not isinstance(node, Const):
            return False
        return node.value is None

    @classmethod
    def get_full_attribute_name(cls, node: Attribute) -> str:
        return cls._get_attribute_value(node)

    @classmethod
    def get_function_name(cls, node: Call) -> str | None:
        if not isinstance(node, Call):
            return None
        if isinstance(node.func, Attribute):
            return node.func.attrname
        if isinstance(node.func, Name):
            return node.func.name
        return None

    @classmethod
    def get_full_function_name(cls, node: Call) -> str | None:
        if not isinstance(node, Call):
            return None
        if isinstance(node.func, Attribute):
            return cls._get_attribute_value(node.func)
        if isinstance(node.func, Name):
            return node.func.name
        return None

    @classmethod
    def _get_attribute_value(cls, node: Attribute):
        if isinstance(node.expr, Name):
            return node.expr.name + '.' + node.attrname
        if isinstance(node.expr, Attribute):
            parent = cls._get_attribute_value(node.expr)
            return node.attrname if parent is None else parent + '.' + node.attrname
        if isinstance(node.expr, Call):
            name = cls.get_full_function_name(node.expr)
            return node.attrname if name is None else name + '.' + node.attrname
        name = type(node.expr).__name__
        if name not in missing_handlers:
            missing_handlers.add(name)
            logger.debug(f"Missing handler for {name}")
        return None


class TreeVisitor:

    def visit(self, node: NodeNG):
        self._visit_specific(node)
        for child in node.get_children():
            self.visit(child)

    def _visit_specific(self, node: NodeNG):
        method_name = "visit_" + type(node).__name__.lower()
        method_slot = getattr(self, method_name, None)
        if callable(method_slot):
            method_slot(node)
            return
        self.visit_nodeng(node)

    def visit_nodeng(self, node: NodeNG):
        pass


class MatchingVisitor(TreeVisitor):

    def __init__(self, node_type: type, match_nodes: list[tuple[str, type]]):
        super()
        self._matched_nodes: list[NodeNG] = []
        self._node_type = node_type
        self._match_nodes = match_nodes

    @property
    def matched_nodes(self):
        return self._matched_nodes

    def visit_assign(self, node: Assign):
        if self._node_type is not Assign:
            return
        self._matched_nodes.append(node)

    def visit_call(self, node: Call):
        if self._node_type is not Call:
            return
        try:
            if self._matches(node.func, 0):
                self._matched_nodes.append(node)
        except NotImplementedError as e:
            logger.warning(f"Missing implementation: {e.args[0]}")

    def visit_import(self, node: Import):
        if self._node_type is not Import:
            return
        self._matched_nodes.append(node)

    def visit_importfrom(self, node: ImportFrom):
        if self._node_type is not ImportFrom:
            return
        self._matched_nodes.append(node)

    def _matches(self, node: NodeNG, depth: int):
        if depth >= len(self._match_nodes):
            return False
        if isinstance(node, Call):
            return self._matches(node.func, depth)
        name, match_node = self._match_nodes[depth]
        if not isinstance(node, match_node):
            return False
        next_node: NodeNG | None = None
        if isinstance(node, Attribute):
            if node.attrname != name:
                return False
            next_node = node.expr
        elif isinstance(node, Name):
            if node.name != name:
                return False
        else:
            raise NotImplementedError(str(type(node)))
        if next_node is None:
            # is this the last node to match ?
            return len(self._match_nodes) - 1 == depth
        return self._matches(next_node, depth + 1)


class NodeBase(ABC):

    def __init__(self, node: NodeNG):
        self._node = node

    @property
    def node(self):
        return self._node

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._node)}>"
